#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <direct.h>
#include <io.h>
#include <fcntl.h>

/*

ODBC Trace File Scanner
=====================
This application scans an ODBC trace log for ODBC handle allocations/releases
and dumps a list of unreleased handles.  It recognizes both ODBC 2.x and 3.x
calls (e.g., SQLAllocEnv, SQLAllocHandle).

In a well-behaved ODBC application, handles are freed when no longer needed.
For statement handles, this is done using SQLFreeStmt( SQL_DROP ) or
SQLFreeHandle. If your application doesn't free handles, it may eventually
run out of handles and encounter a "resource governor exceeded" message.
This program will help you find this type of error.

To use this application, turn on ODBC tracing using the ODBC Administrator
and then run your application.  When the application completes, turn off
tracing.  Use the ODBC trace file scanner to look for unfreed handles.
Here is an example of the program's use including some sample output.

C:\> ODBCSCAN SQL.LOG

Processing SQL.LOG
Unfreed handle 31f1540 allocated at line 10
Unfreed handle 31f21d0 allocated at line 1532
Unfreed handle 320b3f0 allocated at line 2097782 associated with CALL GROUPO.AnalyzeFolder (?,?,?,?)\ 0
Unfreed handle 3203dd0 allocated at line 2098290 associated with CALL GROUPO.InsertBusinessLogic (?,?,?,?)\ 0
Unfreed handle 31f8ca0 allocated at line 2594064 associated with CALL GROUPO.SqlProcess (?,?)\ 0
.
.
.

In the above example, the first two handles are the environment and
connection handles. The program was run up to the point where it ran out of
handles and then the log file was analyzed.  Since the program had not
terminated, the environment and connection handles are still allocated.
The rest of the output contains the important information. It shows which
handles were not freed and at what line in the ODBC trace log the handle
was last allocated.  The SQL statement that is displayed was the last
statement to be prepared or executed directly using the unfreed handle.
This should help you determine where, in the application, the problem
may lie.

The source code for the program is written in C and can be compiled by
either Watcom or Microsoft C compilers. An executable 32-bit Windows
console application is included.

Author
======
Jack Schueler, April 2007
iAnywhere Solutions, Inc.

This program is designed for 32-bit applications and may require some
modification for 64-bit applications with 64-bit handles.

*/

#define ADD 0
#define DEL 1
#define REF 2

enum search_types {
    SEARCHING,
    SEARCHING_HANDLE_ADDRESS,
    SEARCHING_HENV_ADDRESS, 
    SEARCHING_HDBC_ADDRESS,
    SEARCHING_HSTMT_ADDRESS,
    SEARCHING_HANDLE_REF,
    SEARCHING_HENV_REF,
    SEARCHING_HDBC_REF,
    SEARCHING_HSTMT_REF,
    SEARCHING_CHAR_REF,
    SEARCHING_SQL_DROP
};

typedef struct _TARGET_REF {
    struct _TARGET_REF	*next;	    // next reference in list
    unsigned		line;	    // line # where referenced
    unsigned char	*address;    // address
    int			state;	    // allocated or freed
    char		*stmt;	    // pointer to prepared statement
} TARGET_REF;

TARGET_REF *List = NULL;

#define	    MSG_BUFF_SIZE 4096
char	    MsgBuf[MSG_BUFF_SIZE];

char	    LineBuffer[MSG_BUFF_SIZE+256];

void Error( char *msg, int line )
{
    if( line == 0 ) {
	printf( "%s\n", msg );
    } else {
	printf( "%s at line %d\n", msg, line );
    }
    if ( isatty( fileno( stdout ) ) == 0 ) {
	// stdout is redirected
	if( line == 0 ) {
	    fprintf( stderr, "%s\n", msg );
	} else {
	    fprintf( stderr, "%s at line %d\n", msg, line );
	}
    }
    exit( EXIT_FAILURE );
}

void Message( char *msg, int line )
{
    if( line == 0 ) {
	fprintf( stderr, "%s\n", msg );
    } else {
	fprintf( stderr, "%s at line %d\n", msg, line );
    }
}

void LogPrintf( char *format, ... )
{
    va_list         arglist;

    MsgBuf[0] = '\0';

    va_start( arglist, format );

    vsprintf( &MsgBuf[strlen(MsgBuf)], format, arglist );

    printf( "%s", MsgBuf );
    if ( isatty( fileno( stdout ) ) == 0 ) {
	// stdout is redirected
	fprintf( stderr, "%s", MsgBuf );
    }
    
    va_end( arglist );
}

void Lookup( int action, unsigned char *address, unsigned line, char *text )
{
    TARGET_REF	    *list;
    TARGET_REF	    *prev;
    char	    *stmt;

    prev = NULL;
    list = List;
    while( list != NULL ) {
        if( address == list->address ) break;
	prev = list;
        list = list->next;
    }
    if( list == NULL ) {
        list = (TARGET_REF *) malloc( sizeof( TARGET_REF ) );
        if( list == NULL ) {
            Error( "Out of memory!", 0 );
        }
	list->next = NULL;
        list->line = line;
        list->address = address;
	list->state = action;
	list->stmt = NULL;
	if( action == DEL ) {
	    Message( "DEL of non-existent entry", line );
	}
	if( action == REF ) {
	    Message( "REF of non-existent entry", line );
	}
	if( List == NULL ) {
	    List = list;
	} else {
	    prev->next = list;
	}
    } else {
	if( action == DEL && list->state == DEL ) {
	    LogPrintf( "DEL at %d of entry %x deleted at %d\n",
	    	    line, list->address, list->line );
	} else if( action == ADD && list->state == ADD ) {
	    LogPrintf( "ADD at %d of entry %x added at %d\n",
		    line, list->address, list->line );
	}
	if( action == REF ) {
	    if( list->stmt != NULL ) free( list->stmt );
	    list->stmt = malloc( strlen( text ) + 1 );
	    if( list->stmt == NULL ){
		Error( "Out of memory!", 0 );
	    }
	    strcpy( list->stmt, text );
	} else {
	    if( list->stmt != NULL ) free( list->stmt );
	    list->stmt = NULL;
	    list->line = line;
	    list->state = action;
	    // debugging
	    #if 0
	    if( action == ADD ) {
		LogPrintf( "ADD at %d of entry %x\n",
			line, list->address );
	    } else {
		LogPrintf( "DEL at %d of entry %x\n",
			line, list->address );
	    }
	    #endif
	}
    }
}

void DumpList()
{
    TARGET_REF	    *list;
    TARGET_REF	    *prev;
    TARGET_REF	    *next;
    TARGET_REF	    *pprev;

    int		    sorted;

    // Do a simple bubble sort of the list.
    // Sorting by ODBC trace file line numbers.
    sorted = 0;
    while( (!sorted) && (List != NULL) ) {
	sorted = 1;
	pprev = NULL;
	prev = List;
	list = prev->next;
	while( list != NULL ) {
	    if( list->line < prev->line ) {
		if( pprev == NULL ) {
		    List = list;
		    next = list->next;
		    list->next = prev;
		    prev->next = next;
		    sorted = 0;
		} else {
		    pprev->next = list;
		    next = list->next;
		    list->next = prev;
		    prev->next = next;
		    sorted = 0;
		}
		pprev = list;
		list = prev->next;
	    } else {
		pprev = prev;
		prev = list;
		list = list->next;
	    }
	}
    }
    list = List;
    while( list != NULL ) {
        if( list->state == ADD ) {
	    if( list->stmt == NULL ) {
		LogPrintf( "Unfreed handle %x allocated at line %d\n",
			list->address, list->line );
	    } else {
		LogPrintf( "Unfreed handle %x allocated at line %d associated with %s\n",
			list->address, list->line, list->stmt );
	    }
	}
        list = list->next;
    }
}

/*

Typical sorts of things we'll encounter.

application	74c-6f8	EXIT  SQLAllocHandle  with return code 0 (SQL_SUCCESS)
		SQLSMALLINT                  3 <SQL_HANDLE_STMT>
		SQLHANDLE           031F15E8
		SQLHANDLE *         0x00C8E444 ( 0x031f1848)

application     f48-d88 EXIT  SQLAllocStmt  with return code 0 (SQL_SUCCESS)
                HDBC                00000000022D7D90
                HSTMT *             0x000000000023FC60 ( 0x022d5320)
		
application	f4c-6f8 EXIT  SQLPrepareW  with return code 0 (SQL_SUCCESS)
	    	HSTMT               0320B3F0
		WCHAR *             0x09182194 [ -3] "CALL sprocs... (?,?,?,?)\ 0"
		SDWORD              -3

applicaton      b68-dd8 ENTER SQLExecDirect
                HSTMT               05E22390
                UCHAR *             0x653B1C3C [ -3] "SET TEMPORARY OPTION...\ 0"
                SDWORD              -3

application     74c-320	EXIT  SQLFreeHandle  with return code 0 (SQL_SUCCESS)
		SQLSMALLINT                  3 <SQL_HANDLE_STMT>
		SQLHANDLE           031F1848

application     3e8-11f8 EXIT  SQLFreeStmt
		HSTMT               0000000002295540
		UWORD		    1 <SQL_DROP>
		
*/
void SearchHandles( char *file_name )
{
    FILE	*fp;
    int		line = 0;
    int		count;
    int		state;
    char	*p0;
    char	*p1;
    char	*p2;
    char	*address;
       
    LogPrintf( "Processing %s\n", file_name );
    
    fp = fopen( file_name, "r" );

    if( fp == NULL ) Error( "Error - Can't open input file", 0 );


    state = SEARCHING;
    
    while( fgets( LineBuffer, sizeof(LineBuffer) - 1, fp ) != NULL ) {

	line++;
	// LogPrintf( "%d: %s", line, LineBuffer );
	
	p0 = LineBuffer;

	switch( state )
	{
	    case SEARCHING_HANDLE_ADDRESS:
		// SQLHANDLE *         0x00C8E444 ( 0x031f1848)
		if( (p1 = strstr( p0, "SQLHANDLE *" )) != NULL ) {
		    if( (p1 = strstr( p1, "(" )) != NULL ) {
			p1 += 2;
			address = (char *)strtoul( p1, &p2, 16 );
			Lookup( ADD, address, line, "" );
		    } else {
			Error( "Missing HANDLE address", line );
		    }
		    state = SEARCHING;
		    
		}
		break;
	    
	    case SEARCHING_HENV_ADDRESS:
		// HENV *         0x00C8E444 ( 0x031f1848)
		if( (p1 = strstr( p0, "HENV *" )) != NULL ) {
		    if( (p1 = strstr( p1, "(" )) != NULL ) {
			p1 += 2;
			address = (char *)strtoul( p1, &p2, 16 );
			Lookup( ADD, address, line, "" );
		    } else {
			Error( "Missing HANDLE address", line );
		    }
		    state = SEARCHING;
		    
		}
		break;

	    case SEARCHING_HDBC_ADDRESS:
		// HDBC *         0x00C8E444 ( 0x031f1848)
		if( (p1 = strstr( p0, "HDBC *" )) != NULL ) {
		    if( (p1 = strstr( p1, "(" )) != NULL ) {
			p1 += 2;
			address = (char *)strtoul( p1, &p2, 16 );
			Lookup( ADD, address, line, "" );
		    } else {
			Error( "Missing HANDLE address", line );
		    }
		    state = SEARCHING;
		    
		}
		break;
	    
	    case SEARCHING_HSTMT_ADDRESS:
		// HSTMT *         0x00C8E444 ( 0x031f1848)
		if( (p1 = strstr( p0, "HSTMT *" )) != NULL ) {
		    if( (p1 = strstr( p1, "(" )) != NULL ) {
			p1 += 2;
			address = (char *)strtoul( p1, &p2, 16 );
			Lookup( ADD, address, line, "" );
		    } else {
			Error( "Missing HANDLE address", line );
		    }
		    state = SEARCHING;
		    
		}
		break;
	    
	    case SEARCHING_HANDLE_REF:
		// SQLHANDLE           031F1848
		if( (p1 = strstr( p0, "SQLHANDLE" )) != NULL ) {
		    p1 += 9;
		    address = (char *)strtoul( p1, &p2, 16 );
		    Lookup( DEL, address, line, "" );
		    state = SEARCHING;
		}
		break;
	    
	    case SEARCHING_HENV_REF:
		// HENV           031F1848
		if( (p1 = strstr( p0, "HENV" )) != NULL ) {
		    p1 += 4;
		    address = (char *)strtoul( p1, &p2, 16 );
		    Lookup( DEL, address, line, "" );
		    state = SEARCHING;
		}
		break;
		
	    case SEARCHING_HDBC_REF:
		// HSTMT           031F1848
		if( (p1 = strstr( p0, "HDBC" )) != NULL ) {
		    p1 += 4;
		    address = (char *)strtoul( p1, &p2, 16 );
		    Lookup( DEL, address, line, "" );
		    state = SEARCHING;
		}
		break;
		
	    case SEARCHING_SQL_DROP:
		// HSTMT           031F1848
		// UWORD	    1 <SQL_DROP>
		count++;
		if( count > 2 ) {
		    // Give up because we should have seen it by now.
		    // Could have been a SQL_CLOSE.
		    state = SEARCHING;
		} else if( (p1 = strstr( p0, "HSTMT" )) != NULL ) {
		    // save address for later when we see the SQL_DROP
		    p1 += 5;
		    address = (char *)strtoul( p1, &p2, 16 );
		} else if( (p1 = strstr( p0, "<SQL_DROP>" )) != NULL ) {
		    Lookup( DEL, address, line, "" );
		    state = SEARCHING;
		}
		break;

	    case SEARCHING_HSTMT_REF:
		// HSTMT           031F1848
		if( (p1 = strstr( p0, "HSTMT" )) != NULL ) {
		    p1 += 5;
		    address = (char *)strtoul( p1, &p2, 16 );
		    state = SEARCHING_CHAR_REF;
		}
		break;
		
	    case SEARCHING_CHAR_REF:
		// WCHAR *   0x09182194 [ -3] "CALL sprocs... (?,?,?,?)\ 0"
		// UCHAR *   0x653B1C3C [ -3] "SET TEMPORARY OPTION...\ 0"
		if( (p1 = strstr( p0, "CHAR *" )) != NULL ) {
		    if( (p1 = strstr( p1, "\"" )) != NULL ) {
			p1 += 1;
			p2 = strstr( p1, "\"" );
			if( p2 != NULL ) *p2 = '\0';
			Lookup( REF, address, line, p1 );
		    }
		    state = SEARCHING;
		}
		break;
    
	    default:
    
		if( strstr( p0, "EXIT  SQLAllocHandle" ) != NULL ) {
		    if( state != SEARCHING ) {
			Error( "Wrong state", line );
		    }
		    state = SEARCHING_HANDLE_ADDRESS;
		    
		} else if( strstr( p0, "EXIT  SQLAllocEnv" ) != NULL ) {
		    if( state != SEARCHING ) {
			Error( "Wrong state", line );
		    }
		    state = SEARCHING_HENV_ADDRESS;
		    
		} else if( strstr( p0, "EXIT  SQLAllocConnect" ) != NULL ) {
		    if( state != SEARCHING ) {
			Error( "Wrong state", line );
		    }
		    state = SEARCHING_HDBC_ADDRESS;
		    
		} else if( strstr( p0, "EXIT  SQLAllocStmt" ) != NULL ) {
		    if( state != SEARCHING ) {
			Error( "Wrong state", line );
		    }
		    state = SEARCHING_HSTMT_ADDRESS;
		    
		} else if( strstr( p0, "EXIT  SQLFreeHandle" ) != NULL ) {
		    if( state != SEARCHING ) {
			Error( "Wrong state", line );
		    }
		    state = SEARCHING_HANDLE_REF;
		    
		} else if( strstr( p0, "EXIT  SQLFreeEnv" ) != NULL ) {
		    if( state != SEARCHING ) {
			Error( "Wrong state", line );
		    }
		    state = SEARCHING_HENV_REF;
		    
		} else if( strstr( p0, "EXIT  SQLFreeConnect" ) != NULL ) {
		    if( state != SEARCHING ) {
			Error( "Wrong state", line );
		    }
		    state = SEARCHING_HDBC_REF;
		    
		} else if( strstr( p0, "EXIT  SQLFreeStmt" ) != NULL ) {
		    if( state != SEARCHING ) {
			Error( "Wrong state", line );
		    }
		    count = 0;
		    state = SEARCHING_SQL_DROP;
		    
		} else if( strstr( p0, "EXIT  SQLPrepare" ) != NULL ) {
		    if( state != SEARCHING ) {
			Error( "Wrong state", line );
		    }
		    state = SEARCHING_HSTMT_REF;
		    
		} else if( strstr( p0, "EXIT  SQLExecDirect" ) != NULL ) {
		    if( state != SEARCHING ) {
			Error( "Wrong state", line );
		    }
		    state = SEARCHING_HSTMT_REF;
		    
		}
		break;
	}
    }
    fclose( fp );
}

void main( int argc, char *argv[] )
{
    
    if( (( argc <= 1 )                                   ) ||
	(( argc  > 1 ) && (strcmp( argv[1], "-?" ) == 0 )) ) {
	
	Error( "Usage: odbcscan odbc_trace_file_name", 0 );
	
    } else {
	
	SearchHandles( argv[1] );
	DumpList();

    }

}
